from hikariwave.config import (
    validate_bitrate,
    validate_channels,
    validate_volume,
)

import pytest

@pytest.mark.parametrize(
    "input_value,expected", [
        ("6k", "6k"),
        ("96k", "96k"),
        ("510k", "510k"),
        (" 128k ", "128k"),
    ],
)
def test_bitrates_valid(input_value: str, expected: str) -> None:
    assert validate_bitrate(input_value) == expected

@pytest.mark.parametrize(
    "input_value",
    ["5k", "511k", "abc", "100", 100, None],
)
def test_bitrates_invalid(input_value: object) -> None:
    with pytest.raises((TypeError, ValueError)):
        validate_bitrate(input_value)


@pytest.mark.parametrize(
    "input_value,expected",
    [(1, 1), (2, 2)],
)
def test_valid_channels(input_value: int, expected: int):
    assert validate_channels(input_value) == expected

@pytest.mark.parametrize(
    "input_value",
    [0, 3, -1, "2", None],
)
def test_invalid_channels(input_value: object):
    with pytest.raises((ValueError, TypeError)):
        validate_channels(input_value)


@pytest.mark.parametrize(
    "input_value,expected",
    [
        (1.0, 1.0),
        (0.5, 0.5),
        (10, 10.0),
        ("+3dB", "+3dB"),
        ("-2.5dB", "-2.5dB"),
        ("0dB", "0dB"),
    ],
)
def test_valid_volumes(input_value: float | int | str, expected: float | str):
    assert validate_volume(input_value) == expected

@pytest.mark.parametrize(
    "input_value",
    [
        -1.0, -5,
        "3", "dB",
        "3db", "+dB",
        None, [], {},
    ],
)
def test_invalid_volumes(input_value: object):
    with pytest.raises((ValueError, TypeError)):
        validate_volume(input_value)


from __future__ import annotations

from hikariwave.config import (
    validate_bitrate,
    validate_channels,
    validate_volume,
)
from yt_dlp.YoutubeDL import YoutubeDL as YT

import asyncio
import os

__all__ = (
    "AudioSource",
    "BufferAudioSource",
    "FileAudioSource",
    "URLAudioSource",
    "YouTubeAudioSource",
)

def _validate_content(content: object, name: str, expected: tuple[object]) -> type:
    if not isinstance(content, expected):
        if len(expected) == 1:
            error: str = f"Provided {name} must be `{expected}`"
        elif len(expected) == 2:
            error: str = f"Provided {name} must be `{expected[0]}` or `{expected[1]}`"
        else:
            types: list[str] = [f"`{type_.__name__}`" for type_ in expected]
            error: str = f"Provided {name} must be " + ", ".join(types[:-1]) + f", or {types[-1]}"
        
        raise TypeError(error)
    
    try:
        if len(content) == 0:
            error: str = f"Provided {name} can't be empty"
            raise ValueError(error)
    except TypeError:
        pass

    return content

def _validate_name(name: object) -> str:
    if not isinstance(name, str):
        error: str = "Provided name must be `str`"
        raise TypeError(error)
    
    if len(name) == 0:
        error: str = "Provided name cannot be empty"
        raise ValueError(error)
    
    return name

class AudioSource:
    """Base audio source implementation."""

    def __eq__(self, other: object) -> bool:
        if type(self) is not type(other): return False
        return self._content == other._content
    
    def __hash__(self) -> int:
        return hash((type(self), self._content))

    @property
    def bitrate(self) -> str | None:
        """If provided, the bitrate in which this source is played back at."""
        return self._bitrate

    @property
    def channels(self) -> int | None:
        """If provided, the amount of channels this source plays with."""
        return self._channels
    
    @property
    def name(self) -> str | None:
        """If provided, the name assigned to this source for display purposes."""
        return self._name

    @property
    def volume(self) -> float | str | None:
        """If provided, the overriding volume for this source."""
        return self._volume

class BufferAudioSource(AudioSource):
    """Buffered audio source implementation."""

    __slots__ = (
        "_content",
        "_bitrate",
        "_channels",
        "_name",
        "_volume",
    )

    def __init__(
        self,
        buffer: bytearray | bytes | memoryview,
        *,
        bitrate: str | None = None,
        channels: int | None = None,
        name: str | None = None,
        volume: float | str | None = None
    ) -> None:
        """
        Create a buffered audio source.
        
        Parameters
        ----------
        buffer : bytearray | bytes | memoryview
            The audio data as a buffer.
        bitrate : str | None
            If provided, the bitrate in which to play this source back at.
        channels : int | None
            If provided, the amount of channels this source plays with.
        name : str | None
            If provided, an internal name used for display purposes.
        volume : float | str | None
            If provided, overrides the player's set/default volume. Can be scaled (`0.5`, `1.0`, `2.0`, etc.) or dB-based (`-3dB`, etc.).
        
        Raises
        ------
        TypeError
            - If `buffer` is not `bytearray`, `bytes`, or `memoryview`.
            - If `bitrate` is provided and not `str`.
            - If `channels` is provided and not `int`.
            - If `name` is provided and not `str`.
            - If `volume` is provided and not `float` or `str`.
        ValueError
            - If `buffer` is empty.
            - If `bitrate` is provided, and is not between `6k` and `510k`.
            - If `channels` is provided and not `1` or `2`.
            - If `name` is provided and is empty.
            - If `volume` is provided and is either a `float` and is not positive or a `str` and does not end with `dB`, contain a number, or (if provided) doesn't begin with `-` or `+`.
        """

        self._content: bytearray | bytes | memoryview = _validate_content(
            buffer,
            "buffer",
            (bytearray, bytes, memoryview)
        )

        self._bitrate: str | None = validate_bitrate(bitrate) if bitrate is not None else None
        self._channels: int | None = validate_channels(channels) if channels is not None else None
        self._name: str | None = _validate_name(name) if name is not None else None
        self._volume: float | str | None = validate_volume(volume) if volume is not None else None
    
    @property
    def buffer(self) -> bytearray | bytes | memoryview:
        """The audio data as a buffer."""
        return self._content

class FileAudioSource(AudioSource):
    """File audio source implementation."""

    __slots__ = (
        "_content",
        "_bitrate",
        "_channels",
        "_name",
        "_volume",
    )

    def __init__(
        self,
        filepath: str,
        *,
        bitrate: str | None = None,
        channels: int | None = None,
        name: str | None = None,
        volume: float | str | None = None
    ) -> None:
        """
        Create a file audio source.
        
        Parameters
        ----------
        filepath : str
            The filepath to the audio file.
        bitrate : str | None
            If provided, the bitrate in which to play this source back at.
        channels : int | None
            If provided, the amount of channels this source plays with.
        name : str | None
            If provided, an internal name used for display purposes.
        volume : float | str | None
            If provided, overrides the player's set/default volume. Can be scaled (`0.5`, `1.0`, `2.0`, etc.) or dB-based (`-3dB`, etc.).
        
        Raises
        ------
        TypeError
            - If `filepath` is not `str`.
            - If `bitrate` is provided and not `str`.
            - If `channels` is provided and not `int`.
            - If `name` is provided and not `str`.
            - If `volume` is provided and not `float` or `str`.
        ValueError
            - If `filepath` is empty or is not found as a file on the system.
            - If `bitrate` is provided, and is not between `6k` and `510k`.
            - If `channels` is provided and not `1` or `2`.
            - If `name` is provided and is empty.
            - If `volume` is provided and is either a `float` and is not positive or a `str` and does not end with `dB`, contain a number, or (if provided) doesn't begin with `-` or `+`.
        """

        self._content: str = _validate_content(filepath, "filepath", (str,))

        self._bitrate: str | None = validate_bitrate(bitrate) if bitrate is not None else None
        self._channels: int | None = validate_channels(channels) if channels is not None else None
        self._name: str | None = _validate_name(name) if name is not None else None
        self._volume: float | str | None = validate_volume(volume) if volume is not None else None

        if not os.path.isfile(self._content):
            error: str = f"No file exists at this path: {self._content}"
            raise FileNotFoundError(error)
    
    @property
    def filepath(self) -> str:
        """The filepath to the audio file"""
        return self._content

class URLAudioSource(AudioSource):
    """URL audio source implementation."""

    __slots__ = (
        "_content",
        "_bitrate",
        "_channels",
        "_name",
        "_volume",
    )

    def __init__(
        self,
        url: str,
        *,
        bitrate: str | None = None,
        channels: int | None = None,
        name: str | None = None,
        volume: float | str | None = None
    ) -> None:
        """
        Create a URL audio source.
        
        Parameters
        ----------
        url : str
            The URL to the audio source.
        bitrate : str | None
            If provided, the bitrate in which to play this source back at.
        channels : int | None
            If provided, the amount of channels this source plays with.
        name : str | None
            If provided, an internal name used for display purposes.
        volume : float | str | None
            If provided, overrides the player's set/default volume. Can be scaled (`0.5`, `1.0`, `2.0`, etc.) or dB-based (`-3dB`, etc.).
        
        Raises
        ------
        TypeError
            - If `url` is not `str`.
            - If `bitrate` is provided and not `str`.
            - If `channels` is provided and not `int`.
            - If `name` is provided and not `str`.
            - If `volume` is provided and not `float` or `str`.
        ValueError
            - If `url` is empty.
            - If `bitrate` is provided, and is not between `6k` and `510k`.
            - If `channels` is provided and not `1` or `2`.
            - If `name` is provided and is empty.
            - If `volume` is provided and is either a `float` and is not positive or a `str` and does not end with `dB`, contain a number, or (if provided) doesn't begin with `-` or `+`.
        """

        self._content: str = _validate_content(url, "url", (str,))

        self._bitrate: str | None = validate_bitrate(bitrate) if bitrate is not None else None
        self._channels: int | None = validate_channels(channels) if channels is not None else None
        self._name: str | None = _validate_name(name) if name is not None else None
        self._volume: float | str | None = validate_volume(volume) if volume is not None else None

    @property
    def url(self) -> str:
        """The URL to the audio source."""
        return self._content

class YouTubeAudioSource(AudioSource):
    """YouTube audio source implementation."""

    __slots__ = (
        "_url",
        "_bitrate",
        "_channels",
        "_name",
        "_volume",
        "_content",
        "_headers",
        "_future",
    )

    def __init__(
        self,
        url: str,
        *,
        bitrate: str | None = None,
        channels: int | None = None,
        name: str | None = None,
        volume: float | str | None = None
    ) -> None:
        """
        Create a YouTube audio source.
        
        Parameters
        ----------
        url : str
            The YouTube URL of the audio source.
        bitrate : str | None
            If provided, the bitrate in which to play this source back at.
        channels : int | None
            If provided, the amount of channels this source plays with.
        name : str | None
            If provided, an internal name used for display purposes.
        volume : float | str | None
            If provided, overrides the player's set/default volume. Can be scaled (`0.5`, `1.0`, `2.0`, etc.) or dB-based (`-3dB`, etc.).
        
        Important
        ---------
        This source resolves the provided YouTube URL into an internal, direct media URL using `yt-dlp`.
        This resolution is performed asynchronously in the background during construction.

        The resolved media URL may not be immediately available after instantiation. Consumers that require guaranteed availability should `await` the source's completion mechanism (e.g. `await source.wait_for_url()`).

        This source depends on YouTube's undocumented internal APIs via `yt-dlp`. As a result, it is best-effort and may break without notice if YouTube changes its internal behavior.
        Functionality may require updating the pinned `yt-dlp` version to restore compatibility.

        Raises
        ------
        TypeError
            - If `url` is not `str`.
            - If `bitrate` is provided and not `str`.
            - If `channels` is provided and not `int`.
            - If `name` is provided and not `str`.
            - If `volume` is provided and not `float` or `str`.
        ValueError
            - If `url` is empty.
            - If `bitrate` is provided, and is not between `6k` and `510k`.
            - If `channels` is provided and not `1` or `2`.
            - If `name` is provided and is empty.
            - If `volume` is provided and is either a `float` and is not positive or a `str` and does not end with `dB`, contain a number, or (if provided) doesn't begin with `-` or `+`.
        """

        self._url: str = _validate_content(url, "url", (str,))

        self._bitrate: str | None = validate_bitrate(bitrate) if bitrate is not None else None
        self._channels: int | None = validate_channels(channels) if channels is not None else None
        self._name: str | None = _validate_name(name) if name is not None else None
        self._volume: float | str | None = validate_volume(volume) if volume is not None else None

        self._content: str | None = None
        self._headers: dict[str, str] = {}

        loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        self._future: asyncio.Task[tuple[str, dict[str]]] = loop.create_task(self._extract_metadata(loop))

    async def _extract_metadata(self, loop: asyncio.AbstractEventLoop) -> tuple[str, dict[str]]:
        def extract() -> tuple[str, dict[str]]:
            with YT({"quiet": True, "no_warnings": True, "format": "bestaudio[ext=m4a]/bestaudio/best"}) as ydl:
                metadata: dict[str] = ydl.extract_info(self._url, False)
            
                return (
                    metadata["url"],
                    metadata.get("http_headers", {}),
                )

        self._content, self._headers = await loop.run_in_executor(None, extract)
        return self._content, self._headers

    @staticmethod
    def _format_headers(headers: dict[str, str]) -> str:
        return "".join(f"{k}: {v}\r\n" for k, v in headers.items())

    @property
    def future(self) -> asyncio.Task[tuple[str, dict[str]]]:
        """The future that will be completed when the internal media URL is discovered."""
        return self._future

    @property
    def url_media(self) -> str | None:
        """The media source URL that the YouTube URL points to, if discovered - Wait for the source `future` property to finish to attain."""
        return self._content

    @property
    def url_youtube(self) -> str:
        """The URL to the audio source."""
        return self._url
    
    async def wait_for_url(self) -> str:
        """Waits for extraction of the internal media URL, if needed, then returns that URL."""
        if self._content is None:
            await self._future
        
        return self._content


from __future__ import annotations

from hikariwave.internal.constants import Audio
from hikariwave.event.types import VoiceWarningType, WaveEventType
from hikariwave.internal.error import ServerError
from typing import Callable, TYPE_CHECKING

if TYPE_CHECKING:
    from hikariwave.client import VoiceClient, VoiceChannelMeta

import asyncio
import hikari
import logging
import struct

__all__ = ("VoiceServer",)

logger: logging.Logger = logging.getLogger("hikari-wave.server")

class Protocol(asyncio.DatagramProtocol):
    """Background UDP protocol for communication with Discord's voice servers."""

    def __init__(self, ip_discover_future: asyncio.Future[bytes], rtp_listener: Callable[[int], None]) -> None:
        """
        Create a UDP server protocol.
        
        Parameters
        ----------
        ip_discover_future : asyncio.Future[bytes]
            The future to set when our IP is discovered.
        rtp_listener : Callable[[int], None]
            The callback to call when we receive a non-IP discovery packet.
        """
        
        self._transport: asyncio.DatagramTransport = None

        self._ip_discover_future: asyncio.Future[bytes] = ip_discover_future
        self._rtp_listener: Callable[[int], None] = rtp_listener

    def connection_made(self, transport: asyncio.DatagramTransport) -> None:
        """
        Called automatically when a UDP connection is made.
        
        Parameters
        ----------
        transport : asyncio.DatagramTransport
            The UDP transport.
        """
        
        self._transport = transport

    def datagram_received(self, data: bytes, address: tuple[str, int]) -> None:
        """
        Automatically called when we receive a UDP packet.
        
        Parameters
        ----------
        data : bytes
            The UDP packet data
        address : tuple[str, int]
            The address that sent this packet.
        """

        if not self._ip_discover_future.done():
            self._ip_discover_future.set_result(data)
            return

        self._rtp_listener(data)
    
    def error_received(self, exc: Exception):
        """
        Automatically called when an error occurs.
        
        Parameters
        ----------
        exc : Exception
            The error that occurred.
        """
        
        if self._ip_discover_future.done():
            self._ip_discover_future.set_exception(exc)
            return 

class RTPStats:
    def __init__(self) -> None:
        self._prev_arrival: float | None = None
        self._prev_timestamp: int | None = None
        self._jitter: float = 0.0
        self._last_seq: int | None = None
        self._received: int = 0
        self._lost: int = 0
    
    def update(self, seq: int, timestamp: int, arrival_time: float) -> None:
        if self._last_seq is not None:
            expected: int = (self._last_seq + 1) & Audio.BIT_16U
            if seq != expected:
                delta: int = (seq - expected) & Audio.BIT_16U
                self._lost += delta
        
        self._last_seq = seq
        self._received += 1

        arrival: float = arrival_time * Audio.SAMPLING_RATE

        if self._prev_arrival is not None:
            delta: float = (arrival - self._prev_arrival) - (timestamp - self._prev_timestamp)
            self._jitter += (abs(delta) - self._jitter) / 16
        
        self._prev_arrival = arrival
        self._prev_timestamp = timestamp

class VoiceServer:
    """The background server responsible for communicating with Discord's voice servers."""

    __slots__ = (
        "_client", "_ip", "_port", "_ssrc", "_udp", "_last_audio", "_watch_task", "_stats",
    )

    def __init__(
        self,
        client: VoiceClient,
    ) -> None:
        """
        Create a new voice server connection.
        
        Parameters
        ----------
        client : VoiceClient
            The voice client handling all bot connections and state.
        """
        
        self._client: VoiceClient = client

        self._ip: str = None
        self._port: int = None
        self._ssrc: int = None
        self._udp: asyncio.DatagramTransport = None

        self._last_audio: dict[int, float] = {}
        self._watch_task: asyncio.Task = None

        self._stats: dict[int, RTPStats] = {}

    async def _discover_ip(self) -> tuple[str, int]:
        loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        future: asyncio.Future[bytes] = loop.create_future()

        self._udp, _ = await loop.create_datagram_endpoint(
            lambda: Protocol(future, self._rtp_packet),
            remote_addr=(self._ip, self._port),
        )

        packet: bytes = struct.pack("!HHI", 0x1, 70, self._ssrc) + bytes(70)
        self._udp.sendto(packet)

        data: bytes = await asyncio.wait_for(future, 3.0)

        if (type_ := struct.unpack("!H", data[0:2])[0]) != 0x0002:
            error: str = f"Expected packet type 2, got {type_}"
            raise ServerError(error)
    
        if (len_ := struct.unpack("!H", data[2:4])[0]) != 70:
            error: str = f"Expected packet length of 70, got {len_}"
            raise ServerError(error)

        external_ip: str = data[8:72].split(b"\x00", 1)[0].decode("ascii")
        external_port: int = struct.unpack("!H", data[72:74])[0]

        logger.debug(f"External address discovered: ({external_ip}:{external_port})")
        return external_ip, external_port

    def _rtp_packet(self, data: bytes) -> None:
        if len(data) < 12: return

        seq: int = struct.unpack_from(">H", data, 2)[0]
        timestamp: int = struct.unpack_from(">I", data, 4)[0]
        ssrc: int = struct.unpack_from(">I", data, 8)[0]

        if ssrc not in self._client._ssrcsr:
            return

        now: float = asyncio.get_running_loop().time()

        stats: RTPStats | None = self._stats.get(ssrc)
        if stats is None:
            stats = self._stats[ssrc] = RTPStats()
        
        stats.update(seq, timestamp, now)

        is_new: bool = ssrc not in self._last_audio
        self._last_audio[ssrc] = now

        if not is_new:
            return
        
        user_id: hikari.Snowflake = self._client._ssrcsr[ssrc]
        channel_id: hikari.Snowflake = self._client._members[user_id]

        channel: VoiceChannelMeta = self._client._channels[channel_id]
        guild: hikari.Snowflake = channel.guild_id

        self._client._event_factory.emit(
            WaveEventType.MEMBER_START_SPEAKING,
            channel_id,
            guild,
            channel.members[user_id],
        )
        
        channel.active.add(user_id)

    async def _watch_silence(self) -> None:
        loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()

        try:
            while True:
                now: float = loop.time()

                for ssrc, last in list(self._last_audio.items()):
                    if now - last <= 0.25:
                        continue

                    del self._last_audio[ssrc]
                    self._stats.pop(ssrc, None)

                    user_id: hikari.Snowflake = self._client._ssrcsr[ssrc]
                    channel_id: hikari.Snowflake = self._client._members[user_id]

                    channel: VoiceChannelMeta = self._client._channels[channel_id]
                    guild: hikari.Snowflake = channel.guild_id

                    channel.active.remove(user_id)

                    self._client._event_factory.emit(
                        WaveEventType.MEMBER_STOP_SPEAKING,
                        channel,
                        guild,
                        channel.members[user_id],
                    )
                
                for ssrc, stats in self._stats.items():
                    if stats is None:
                        continue

                    jitter: float = stats._jitter / (Audio.SAMPLING_RATE // 1000)
                    total: int = stats._received + stats._lost
                    loss_rate: float = (stats._lost / total) if total else 0.0

                    user_id: hikari.Snowflake = None
                    channel_id: hikari.Snowflake = None
                    channel: VoiceChannelMeta = None
                    guild_id: hikari.Snowflake = None

                    if jitter > Audio.MAX_JITTER:
                        user_id = self._client._ssrcsr[ssrc]
                        channel_id = self._client._members[user_id]
                        channel = self._client._channels[channel_id]
                        guild_id = channel.guild_id

                        self._client._event_factory.emit(
                            WaveEventType.VOICE_WARNING,
                            channel_id,
                            guild_id,
                            VoiceWarningType.JITTER,
                            jitter,
                        )
                    
                    if loss_rate > Audio.MAX_PACKET_LOSS:
                        user_id = self._client._ssrcsr[ssrc]
                        channel_id = self._client._members[user_id]
                        channel = self._client._channels[channel_id]
                        guild_id = channel.guild_id

                        self._client._event_factory.emit(
                            WaveEventType.VOICE_WARNING,
                            channel_id,
                            guild_id,
                            VoiceWarningType.PACKET_LOSS,
                            loss_rate,
                        )

                await asyncio.sleep(0.05)
        except asyncio.CancelledError:
            return

    async def connect(self, ip: str, port: int, ssrc: int) -> tuple[str, int]:
        """
        Connect to a Discord voice server.
        
        Parameters
        ----------
        ip : str
            The Discord voice server's IP.
        port : int
            The Discord voice server's port.
        ssrc : int
            Our assigned SSRC by Discord's voice gateway.
        
        Returns
        -------
        tuple[str, int]
            Our public, discovered IP address.
        """
        
        if self._udp:
            return

        self._ip, self._port, self._ssrc = ip, port, ssrc

        logger.debug(f"Connecting to voice server: IP={ip}, Port={port}, SSRC={ssrc}")
        local_address: tuple[str, int] = await self._discover_ip()

        self._watch_task = asyncio.create_task(self._watch_silence())

        return local_address

    async def disconnect(self) -> None:
        """
        Disconnect from Discord's voice server.
        """
        
        logger.debug(f"Disconnected from server: IP={self._ip}, Port={self._port}")
        
        if self._watch_task:
            self._watch_task.cancel()
            self._watch_task = None
        
        self._last_audio.clear()
        self._stats.clear()

        if self._udp:
            self._udp.close()
            self._udp = None
    
    async def send(self, data: bytes) -> None:
        """
        Send a UDP packet to Discord's voice server.
        
        Parameters
        ----------
        data : bytes
            The UDP packet to send.
        """
        
        if not self._udp or self._udp.is_closing():
            return
    
        self._udp.sendto(data)